#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Résolution de sudoku."""

# Recodé le 12/03/13, parce que je retrouvais pas l'ancien
# … et que j'en avais marre du projet why3

import sudoku_types

VERBOSE = False

DEPTH = 0

# Fonctions logiques
def exists(l):
    """Vérifie qu'au moins une valeur de l est vraie."""
    return not all([not i for i in l])

def one(l):
    """Vérifie que l contient une et une seule valeur de vérité."""
    return sum([bool(i) for i in l]) == 1

def none(l):
    """Vérifie que l ne contient aucune valeur de vérité."""
    return all([not i for i in l])

class CheckError(Exception):
    """Levée si une grille est insoluble."""
    def __init__(self, msg):
        Exception.__init__(self, msg)

def quick_input():
    """Lit une chaîne et renvoie un les caractères un par un. Ne marche que pour le sudoku à caractère unique"""
    chain = raw_input("Input (_=vide):")
    # On remplace quand même les 0 par des _
    chain = chain.replace("0", "_")
    return [i for i in chain]

class Grid(object):
    """Grille de sudoku, contenant des valeurs."""
    
    def check_depth(self):
        """Affiche la valeur de la profondeur d'hypothèse si on n'a jamais atteint ce niveau"""
        global DEPTH
        if self.depth > DEPTH:
            # On affiche l'état avant de faire l'hypothèse
            print self
            print "Profondeur : %s" % self.depth
            DEPTH += 1
    
    def __init__(self, typ, values, depth=0):
        if len(values) != typ.size:
            raise ValueError("%s valeurs incompatibles avec le type de sudoku (%s)" % (len(values), typ.size))
        self.typ = typ
        self.values = [i for i in values] # On fait une copie de la liste pour avoir un passage par valeurs
        self.possible = [[True for i in range(len(typ.chars))] for j in range(len(values))]
        self.depth = depth
        self.check_depth()
    
    def case_filled(self, i):
        """Dit si la case ``i`` contient une valeur."""
        return self.values[i] in self.typ.chars
    
    def deduce(self):
        """Rajoute des valeurs dans les cases en déduisant de ``self.possible``.
           On renvoie ``True`` si on a fait quelque chose."""
        something = False
        # Si une case ne peut plus contenir qu'un seul symbole, on le met dedans
        for case in range(self.typ.size):
            if not self.case_filled(case) and one(self.possible[case]):
                place = self.possible[case].index(True)
                self.values[case] = self.typ.chars[place]
                if VERBOSE:
                    print "%s placé en %s [case]" % (self.typ.chars[place], place)
                something = True
        # Si un linked_field contient déjà tous les symboles sauf un, on le place dans la case restante
        for field in self.typ.linked_fields:
            for nchar in range(len(self.typ.chars)):
                l_field = [self.possible[i][nchar] for i in field]
                if one(l_field):
                    place = field[l_field.index(True)]
                    if not self.case_filled(place):
                        self.values[place] = self.typ.chars[nchar]
                        if VERBOSE:
                            print "%s placé en %s [field %s]" % (self.typ.chars[nchar], place, field)
                        something = True
        return something
    
    def __str__(self):
        """Affiche la grille."""
        return self.typ.render(self.values)
    
    def update_possible(self):
        """Enlève des possibilités en déduisant de ``self.values``."""
        for case in range(self.typ.size):
            if self.case_filled(case): # la case contient quelque chose
                numchar = self.typ.chars.index(self.values[case])
                # Il n'y a plus rien de possible dans cette case
                for p in range(len(self.typ.chars)):
                    self.possible[case][p] = False
                # Il est impossible que la valeur soit dans tout linked_field de cette case
                for f in self.typ.linked_fields:
                    if case in f:
                        for c in f:
                            self.possible[c][numchar] = False
    
    def is_full(self):
        """Dit si la grille est pleine."""
        return all([self.case_filled(i) for i in range(self.typ.size)])
    
    def logic_solve(self):
        """Résoud la grille uniquement par déduction.
           Renvoie ``True`` si elle est entièrement résolue.
           """
        action = True
        while action:
            self.update_possible()
            action = self.deduce()
        return self.is_full()
    
    def _first_empty(self):
        """Renvoie l'index de la première case vide."""
        for i in range(self.typ.size):
            if not self.case_filled(i):
                return i
    
    def hypothesis(self):
        """Émet une hypothèse."""
        if self.is_full():
            raise RuntimeError("Pas besoin de faire une hypothèse, la grille est pleine !")
        # On prend la première case vide
        where = self._first_empty()
        # On choisi le premier symbole possible
        what = self.possible[where].index(True)
        return where, what
    
    def check(self):
        """Vérifie qu'une grille est toujours cohérente."""
        # Les cases dans lesquelles on ne peut plus rien mettre
        empty_cases = [none(self.possible[c]) for c in range(self.typ.size)]
        for c in range(self.typ.size):
            if empty_cases[c] and not self.case_filled(c):
                # La case ne peut rien contenir et ne contient rien
                raise CheckError("La case %s ne peut rien contenir !" % (c))
        # Les linked_field dans lesquels un symbole ne peut plus être placé
        for nchar in range(len(self.typ.chars)):
            empty_fields = [none([self.possible[c][nchar] for c in f]) for f in self.typ.linked_fields]
            for nfield in range(len(self.typ.linked_fields)):
                field = self.typ.linked_fields[nfield]
                lll = [self.values[c] == self.typ.chars[nchar] for c in field]
                if empty_fields[nfield] and not exists(lll):
                    # Le field ne peut pas contenir le caractère et ne le contient effectivement pas
                    raise CheckError("Le field %s ne peut pas contenir de %s !" % (field, self.typ.chars[nchar]))
    
    def is_valid(self):
        """Répond ``True`` si la grille est cohérente."""
        try:
            self.check()
            return True
        except CheckError:
            return False
    
    def hypothetic_solve(self):
        """Résoud en faisant des hypothèses.
           ``self.update_possible`` doit avoir été appliqué."""
        where, what = self.hypothesis()
        charwhat = self.typ.chars[what]
        daughter = Grid(self.typ, self.values, self.depth+1)
        # On applique l'hypothèse
        if VERBOSE:
            print "Hypothèse : %s en %s." % (charwhat, where)
        daughter.values[where] = charwhat
        try:
            solved = daughter.solve()
        except CheckError as e:
            # L'hypothèse était fausse
            if VERBOSE:
                print e.message
                print "L'hypothèse était fausse"
            self.possible[where][what] = False
            return self.solve()
        if solved:
            self.values = daughter.values
            return True
        else:
            raw_input("WTF")
    
    def solve(self):
        """Résoud la grille."""
        solved = self.logic_solve()
        if solved:
            return True
        else:
            self.check()
            return self.hypothetic_solve()
    




if __name__ == "__main__":
    basic = sudoku_types.BasicSudoku()
    samurai = sudoku_types.Samurai()
    #samurai.debug(True)
    #values = [i for i in sudoku_types.examples.hard]
    #values = [i for i in "0"*369]
    values = quick_input()
    #values = [i for i in sudoku_types.examples.samurai]
    g = Grid(basic, values)
    print g
    raw_input("\ngo ?")
    g.solve()
    print g
